"""
Rewrite list types to tuple or scalar types depending on how they are used.
"""
from dataclasses import dataclass, field
from typing import List, cast

from relationalai.semantics.metamodel import ir, visitor, compiler, builtins as bt
from relationalai.semantics.metamodel.util import ordered_set

@dataclass
class RewriteListTypes(compiler.Pass):
    """
    Rewrite list types to tuple or scalar types depending on how they are used.
    """
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        v = RewriteListTypesVisitor()
        result = v.walk(model)
        return result

@dataclass
class RewriteListTypesVisitor(visitor.Rewriter):
    """
    A pass that fixes the types of nodes that use ListTypes.
    """
    new_relations: List[ir.Relation] = field(default_factory=list, init=False)
    new_types: List[ir.Type] = field(default_factory=list, init=False)

    def handle_model(self, model: ir.Model, parent: None):
        result = super().handle_model(model, parent)
        relations_without_list_type = ordered_set(*[
            r for r in result.relations if not any(isinstance(f.type, ir.ListType) for f in r.fields)
        ]).frozen()
        return model.reconstruct(
            result.engines,
            relations_without_list_type | self.new_relations,
            result.types | self.new_types,
            result.root,
            result.annotations,
        )

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
        if any(isinstance(f.type, ir.ListType) for f in node.relation.fields):
            return self._rewrite_non_aggr_relation(node, node.relation)
        else:
            return node

    def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node):
        new_aggr = self._rewrite_aggr_relation(node, node.aggregation)
        return ir.Aggregate(
            node.engine,
            new_aggr,
            node.projection,
            node.group,
            node.args,
        )

    def _rewrite_aggr_relation(self, node: ir.Aggregate, relation: ir.Relation) -> ir.Relation:
        assert len(relation.fields) >= 1
        changed = False

        if changed:
            overloads = ordered_set(*[
                self._rewrite_aggr_relation(node, overload)
                for overload in relation.overloads
            ]).frozen()

            new_relation = ir.Relation(
                relation.name,
                tuple(relation.fields),
                relation.requires,
                relation.annotations,
                overloads,
            )
            self.new_relations.append(new_relation)
            return new_relation

        return relation

    def _rewrite_non_aggr_relation(self, node: ir.Lookup, relation: ir.Relation):
        # Handle the simple varargs case: one list field and all the rest are scalar.
        # Currently this pattern is only used for `rel_primitive_hash_tuple`.
        # and `rel_primitive_solverlib_fo_appl`.
        list_field_indexes = [i for i in range(len(node.relation.fields)) if isinstance(node.relation.fields[i].type, ir.ListType)]
        if len(list_field_indexes) != 1:
            return node

        # There exactly one list field, rewrite it to a tuple type.
        i = list_field_indexes[0]
        assert isinstance(node.relation.fields[i].type, ir.ListType)
        scalar_field_count = len(node.relation.fields) - 1
        tuple_len = len(node.args) - scalar_field_count
        assert tuple_len >= 0, f"List field {i} has {tuple_len} elements, but there are only {scalar_field_count} scalar fields."

        # Flatten the list field into separate scalar fields.
        field = node.relation.fields[i]
        ft = cast(ir.ListType, field.type)
        new_fields = (
            list(node.relation.fields[0:i]) +
            [ir.Field(f"{field.name}@{j}", ft.element_type, field.input) for j in range(tuple_len)] +
            list(node.relation.fields[i+1:])
        )

        # Create new relation, adding external annotation to prevent renaming.
        # NOTE(coey): maybe this is a misuse of this annotation and there could be a nicer way to do it with overloads.
        annos = ordered_set(*node.relation.annotations, bt.external_annotation).frozen()
        new_relation = ir.Relation(node.relation.name, tuple(new_fields), node.relation.requires, annos)
        self.new_relations.append(new_relation)

        return ir.Lookup(node.engine, new_relation, node.args)



from __future__ import annotations
from typing import Any

from relationalai.semantics.internal import internal as b

# Coerce a number to Int64.
def int64(value: b.Producer|int) -> b.ConceptMember:
    return b.ConceptMember(b.Int64, value, {})

# Coerce a number to Int128.
def int128(value: b.Producer|int) -> b.ConceptMember:
    return b.ConceptMember(b.Int128, value, {})

def _make_expr(op: str, *args: Any) -> b.Expression:
    return b.Expression(b.Relationship.builtins[op], *args)

def parse_int64(value: b.Producer|str) -> b.Expression:
    return _make_expr("parse_int64", value, b.Int64.ref("res"))

def parse_int128(value: b.Producer|str) -> b.Expression:
    return _make_expr("parse_int128", value, b.Int128.ref("res"))

# Alias parse_int128 to parse
def parse(value: b.Producer|str) -> b.Expression:
    return parse_int128(value)


import warnings

from relationalai.semantics.reasoners.graph.paths.api import path, star, match, node, union, node_filter

__all__ = ["path", "star", "match", "node", "union", "node_filter"]

warnings.warn(
    "relationalai.early_access.paths.api is deprecated. "
    "Please migrate to relationalai.semantics.reasoners.graph.paths.api",
    DeprecationWarning,
    stacklevel=2,
)


from relationalai.semantics.metamodel.rewrite import Splinter, RewriteListTypes, GarbageCollectNodes, Flatten, \
    DNFUnionSplitter, ExtractKeys, ExtractNestedLogicals, FDConstraints, flatten

__all__ = ["Splinter", "RewriteListTypes", "GarbageCollectNodes", "Flatten", "DNFUnionSplitter", "ExtractKeys",
           "ExtractNestedLogicals", "FDConstraints", "flatten"]



# __init__.py for rpq module
import warnings

from relationalai.semantics.reasoners.graph.paths.benchmarks.grid_graph import create_labeled_grid

__all__ = ["create_labeled_grid"]

warnings.warn(
    "relationalai.early_access.paths.benchmarks is deprecated. "
    "Please migrate to relationalai.semantics.reasoners.graph.paths.benchmarks",
    DeprecationWarning,
    stacklevel=2,
)



from __future__ import annotations

from relationalai.semantics.metamodel import ir
from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel.visitor import Visitor, Rewriter
from relationalai.semantics.metamodel.util import OrderedSet
from relationalai.semantics.metamodel import helpers, factory as f, types, builtins
from typing import Optional, Any, Iterable
from collections import defaultdict

"""
Given an Output with a group of keys (some of them potentially null),
* extract the lookups that bind (transitively) all the keys
* extract the lookups that bind (transitively) properties of the keys
* generate all the valid combinations of keys being present or not
  * first all keys are present,
  * then we remove one key at a time,
  * then we remove two keys at a time,and so on.
  * the last combination is when all the *nullable* keys are missing.
* for each combination:
  * create a compound (hash) key
  * create a Logical that:
      * contains all the relevant lookups for keys and properties
      * contains negated lookups for null keys
      * outputs using the compound key

E.g., we go from

Logical
    Foo(foo)
    rel1(foo, x)
    Logical ^[v1=None]
        rel2(foo, v1)
    Logical ^[v2=None, k2=None]
        rel3(foo, k2)
        rel4(k2, v2)
    Logical ^[v3=None, k3=None]
        rel5(foo, y)
        rel6(y, k3)
        rel7(k3, v3)
    output[foo, k2, k3](v1, v2, v3)

to

Logical
    Logical
        Logical
            Foo(foo)
            rel1(foo, x)
            Logical ^[k2=None, k3=None]
                rel3(foo, k2)
                rel5(foo, y)
                rel6(y, k3)
            Logical ^[v1=None]
                rel2(foo, v1)
            Logical ^[v2=None, k2=None]
                rel4(k2, v2)
            Logical ^[v3=None, k3=None]
                rel7(k3, v3)
            construct(Hash, "Foo", foo, "Concept2", k2, "Concept3", k3, compound_key)
            output[compound_key](v1, v2, v3)
        Logical
            Foo(foo)
            rel1(foo, x)
            Logical ^[k2=None, k3=None]
                rel5(foo, y)
                rel6(y, k3)
            Not
                Logical
                    rel3(foo, k2)
            Logical ^[v1=None]
                rel2(foo, v1)
            Logical ^[v3=None, k3=None]
                rel7(k3, v3)
            construct(Hash, "Foo", foo, "Concept3", k3, compound_key)
            output[compound_key](v1, None, v3)
        Logical
            Foo(foo)
            rel1(foo, x)
            Logical ^[k2=None, k3=None]
                rel3(foo, k2)
                rel5(foo, y)
            Not
                Logical
                    rel6(y, k3)
            Logical ^[v1=None]
                rel2(foo, v1)
            Logical ^[v2=None, k2=None]
                rel4(k2, v2)
            construct(Hash, "Foo", foo, "Concept2", k2, compound_key)
            output[compound_key](v1, v2, None)
        Logical
            Foo(foo)
            rel1(foo, x)
            Logical ^[k2=None, k3=None]
                rel5(foo, y)
            Not
                Logical
                    rel3(foo, k2)
                    rel6(y, k3)
            Logical ^[v1=None]
                rel2(foo, v1)
            construct(Hash, "Foo", foo, compound_key)
            output[compound_key](v1, None, None)
"""
class ExtractKeys(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        visitor = IdentifyKeysVisitor()
        model.accept(visitor)
        return ExtractKeysRewriter(visitor).walk(model) if visitor.extract_info_for_logical else model

class ExtractInfo:
    def __init__(self, vars_to_extract: OrderedSet[ir.Var]):
        self.original_keys = vars_to_extract
        # the subsets of the original keys that are nullable or not.
        self.non_nullable_keys: OrderedSet[ir.Var] = OrderedSet()
        self.nullable_keys: OrderedSet[ir.Var] = OrderedSet()

        # lookup tasks that transitively bind all the key vars
        # e.g., if the key is Z, in foo(X, Y), bar(Y, Z), baz(Z, W)
        # we extract foo(X, Y), bar(Y, Z)
        self.key_lookups: list[ir.Task] = []

        # lookup tasks that transitively bind all the properties of the keys
        # e.g., if the key is Z, in bar(Y, Z), baz(Z, W), qux(W, V)
        # we extract baz(Z, W), qux(W, V).
        # store the lookups by key. each key might bind multiple properties, each list in
        # dict keeps track of one of these properties
        self.inner_property_lookups: dict[ir.Var, list[OrderedSet[ir.Task]]] = defaultdict(list)

        # other top-level tasks that are not key lookups. These need to be kept in each
        # match case of the rewritten Logical.
        self.non_key_top_level_tasks: list[ir.Task] = []

        # the original output task. New outputs are generated based on the keys and
        # properties in the original output
        self.original_output: Optional[ir.Output] = None

        # variables coming from concept lookups, that are not nullable
        # this includes lookups at the top-level, as well as inside logicals when
        # hoisted without a None default.
        self.non_nullable_vars: OrderedSet[ir.Var] = OrderedSet()

class IdentifyKeysVisitor(Visitor):
    def __init__(self):
        self.extract_info_for_logical: dict[ir.Logical, ExtractInfo] = {}
        self.curr_info = None

    def enter(self, node: ir.Node, parent: Optional[ir.Node]=None) -> Visitor:
        if isinstance(node, ir.Logical):
            outputs = [x for x in node.body if isinstance(x, ir.Output) and x.keys]
            if not outputs:
                return self
            assert len(outputs) == 1, "multiple outputs with keys in a logical"
            if not outputs[0].keys:
                return self

            # Logical with an output that has keys
            info = ExtractInfo(OrderedSet.from_iterable(outputs[0].keys))

            # Set the original output
            info.original_output = outputs[0]

            # The original keys and any intermediate vars needed to correctly bind the keys.
            # Any transitive key vars are added during the `find_key_lookups_fixpoint` calls
            extended_keys:OrderedSet[ir.Var] = OrderedSet.from_iterable(outputs[0].keys)

            # First, collect all the top-level lookups
            top_level_lookups = []
            for task in node.body:
                if isinstance(task, ir.Lookup):
                    top_level_lookups.append(task)

            # Find all the key lookups at the top level. Property lookups are handled later
            # and stored in non_key_top_level_tasks
            key_lookups = self.find_key_lookups_fixpoint(top_level_lookups, extended_keys)

            # Then, deal with lookups inside logicals (with hoisted defaults)
            for task in node.body:
                processed = False

                # Top-level concept lookups identify variables that are not nullable
                if isinstance(task, ir.Lookup) and helpers.is_concept_lookup(task):
                    vars = helpers.vars(task.args)
                    if vars[0] not in info.non_nullable_vars:
                        info.non_nullable_vars.add(vars[0])

                if isinstance(task, ir.Logical):
                    for h in task.hoisted:
                        # Hoisted vars without a default are not nullable
                        if isinstance(h, ir.Var):
                            info.non_nullable_vars.add(h)
                        elif isinstance(h, ir.Default):
                            # Hoisted vars with a non-None default are not nullable
                            if h.value is not None:
                                info.non_nullable_vars.add(h.var)
                            elif h.var in info.original_keys and h.var not in info.non_nullable_vars:
                                info.nullable_keys.add(h.var)

                    # Find key lookups inside the logical
                    current_lookups = self.find_key_lookups_fixpoint(task.body, extended_keys)
                    key_lookups.update(current_lookups)

                    # Find property lookups inside the logical. Assume that each logical
                    # binds a separate property for each key
                    current_property_lookups = self.find_property_lookups_fixpoint(task.body, extended_keys)
                    for k, v in current_property_lookups.items():
                        info.inner_property_lookups[k].append(v)

                    if current_lookups or current_property_lookups:
                        processed = True

                if task in key_lookups:
                    continue

                # If the task does not contain key or property lookups, then we need to
                # maintain these at the top level during the actual rewrite
                if not processed and \
                    (
                        isinstance(task, ir.Logical) or \
                        isinstance(task, ir.Lookup) or \
                        isinstance(task, ir.Construct) or \
                        isinstance(task, ir.Aggregate) or \
                        isinstance(task, ir.Rank)
                    ):
                    info.non_key_top_level_tasks.append(task)

            info.non_nullable_keys = info.original_keys - info.nullable_keys
            info.key_lookups = list(key_lookups)

            # we only need to transform the logical if there are nullable keys
            if info.nullable_keys:
                self.extract_info_for_logical[node] = info
                self.curr_info = info
        return self

    def leave(self, node: ir.Node, parent: Optional[ir.Node]=None) -> ir.Node:
        if not self.curr_info:
            return node

        if isinstance(node, ir.Aggregate):
            # we assume that variables appearing in aggregate group-by's are not nullable
            for v in node.group:
                if v in self.curr_info.nullable_keys:
                    self.curr_info.nullable_keys.remove(v)
        elif isinstance(node, ir.Logical) and node in self.extract_info_for_logical:
            # if the set of nullable keys became empty, we shouldn't attempt to transform the logical
            if not self.curr_info.nullable_keys:
                self.extract_info_for_logical.pop(node)
            self.curr_info = None

        return node

    def find_key_lookups_fixpoint(self, tasks:Iterable[ir.Task], keys:OrderedSet[ir.Var]):
        # lookups with a single argument correspond to concepts.
        # we should keep them ahead of the other lookups.
        concept_lookups = OrderedSet()
        # for lookups with multiple arguments, we start from those that have a key as the last
        # argument and move backwards. that's why each time we insert at the front of the list
        lookups = OrderedSet()

        there_is_progress = True
        while there_is_progress:
            there_is_progress = False
            for task in tasks:
                if isinstance(task, ir.Lookup) and task not in lookups and task not in concept_lookups:
                    vars = helpers.vars(task.args)
                    if len(vars) == 1 and vars[0] in keys:
                        concept_lookups.add(task)
                        there_is_progress = True
                    elif len(vars) > 1 and all(v in keys for v in vars[1:]):
                        assert isinstance(vars[0], ir.Var)
                        keys.add(vars[0])
                        lookups.prepend(task)
                        there_is_progress = True

        return concept_lookups | lookups

    def find_property_lookups_fixpoint(self, tasks:Iterable[ir.Task], keys:Iterable[ir.Var]):
        property_lookups = defaultdict(OrderedSet)

        # Keep track of transitive vars seen for each key
        seen_vars: dict[ir.Var, set[ir.Var]] = {k: set([k]) for k in keys}

        there_is_progress = True
        while there_is_progress:
            there_is_progress = False
            for task in tasks:
                if isinstance(task, ir.Lookup):
                    vars = task.args
                    if len(vars) <= 1:
                        # Singleton lookups cannot be for properties
                        continue
                    if any(task in v for v in property_lookups.values()):
                        # Already seen this lookup - don't add it again
                        continue
                    for (var, inter_vars) in seen_vars.items():
                        if any(nv in vars for nv in inter_vars):
                            # If there are any other keys in the lookup, then it is not
                            # purely a lookup of this key's properties
                            #
                            # TODO: handle the case where there are multiple keys for a
                            # single property.
                            if any(isinstance(v, ir.Var) and v in keys and v != var for v in vars):
                                continue

                            # If any seen var is in the args
                            property_lookups[var].add(task)
                            seen_vars[var].update([v for v in vars if isinstance(v, ir.Var)])
                            there_is_progress = True
                elif isinstance(task, ir.Construct):
                    if any(task in v for v in property_lookups.values()):
                        continue
                    for (var, inter_vars) in seen_vars.items():
                        if task.id_var in inter_vars:
                            property_lookups[var].add(task)
                            there_is_progress = True

        return property_lookups

class ExtractKeysRewriter(Rewriter):
    def __init__(self, visitor: IdentifyKeysVisitor):
        super().__init__()
        self.visitor = visitor

    def handle_logical(self, node: ir.Logical, parent: ir.Node, ctx:Optional[Any]=None) -> ir.Logical:
        new_body = self.walk_list(node.body, node)

        # We are in a logical with an output at this level.
        if node in self.visitor.extract_info_for_logical:
            info = self.visitor.extract_info_for_logical[node]

            # Update the key sets based on the identified non-nullable vars
            # remove nullable keys that were inferred to be non-nullable,
            # and add them to the non-nullable set
            non_null_nullable_keys = info.nullable_keys & info.non_nullable_vars
            info.non_nullable_keys.update(non_null_nullable_keys)
            info.nullable_keys = info.nullable_keys - info.non_nullable_vars

            # Get the key lookups. Each output case will start with these lookups
            top_level_key_lookups = list(info.key_lookups)

            # Create a compound key that will be used in place of the original keys.
            compound_key = f.var("compound_key", types.Hash)

            # Get all cases for each combination of null/non-null keys
            match_cases = self._nullable_key_combinations(info, top_level_key_lookups, compound_key)

            return f.logical(tuple(match_cases), [])
        else:
            return node if new_body is node.body else f.logical(new_body, node.hoisted)

    # Generate all the combinations of nullable keys being present or not.
    # For each such combination, generate a list of lookups that will be used in the Match.
    # In total, return a list of lists of lookups, covering all the Match cases.
    def _nullable_key_combinations(
            self,
            info:ExtractInfo,
            inner_key_lookups:list[ir.Task],
            compound_key:ir.Var):
        return self._nullable_key_combinations_rec(info, [], 0, inner_key_lookups, compound_key)

    def _nullable_key_combinations_rec(
            self,
            info: ExtractInfo,
            nullable_non_null_keys: list[ir.Var],
            idx: int,
            inner_key_lookups: list[ir.Task],
            compound_key:ir.Var):

        if idx < len(info.nullable_keys):
            key = info.nullable_keys[idx]
            case1_list = self._nullable_key_combinations_rec(info, nullable_non_null_keys + [key], idx + 1, inner_key_lookups, compound_key)
            case2_list = self._nullable_key_combinations_rec(info, nullable_non_null_keys, idx + 1, inner_key_lookups, compound_key)
            case1_list.extend(case2_list)
            return case1_list
        else:
            # Create a copy to mutate
            inner_key_lookups_ = list(inner_key_lookups)

            # Make sure the orig output is set
            assert isinstance(info.original_output, ir.Output)

            curr_non_null_keys = OrderedSet[ir.Var]()
            curr_null_keys = OrderedSet[ir.Var]()
            for key in info.original_keys:
                if key in info.non_nullable_keys or key in nullable_non_null_keys:
                    curr_non_null_keys.add(key)
                else:
                    curr_null_keys.add(key)

            removed_lookups = self._remove_from_key_lookups(inner_key_lookups_, curr_null_keys, curr_non_null_keys)

            # Build the inner lookups.
            # Start with the non-null key lookups. Wrap these in a logical where the non-null
            # keys are hoisted
            inner_lookups: list[ir.Task] = [f.logical(inner_key_lookups_, list(curr_non_null_keys))]

            # Add other top-level tasks that are not key or property lookups
            inner_lookups.extend(info.non_key_top_level_tasks)

            # Add negated lookups for each null key
            if removed_lookups:
                for (_, removed) in removed_lookups.items():
                    if len(removed) == 1:
                        negated_null_key_lookups = f.not_(removed[0])
                    else:
                        negated_null_key_lookups = f.not_(f.logical(removed, []))
                    inner_lookups.append(negated_null_key_lookups)

            # Add property lookups for each non-null key
            for key in curr_non_null_keys:
                for tasks in info.inner_property_lookups.get(key, []):
                    # Create a separate logical for each key's property lookups. This is
                    # needed so that flatten can split the dependencies correctly
                    prop_lookups = []
                    hoisted_prop_vars = OrderedSet()
                    for task in tasks:
                        if task not in prop_lookups:
                            prop_lookups.append(task)
                            assert isinstance(task, ir.Lookup)

                            # only care about hoisting properties that are in the output
                            vars = helpers.vars(task.args)
                            output_vars = helpers.output_vars(info.original_output.aliases)
                            hoisted_prop_vars.update([v for v in vars if v in output_vars])
                    if prop_lookups:
                        hoisted_props = [ir.Default(v, None) for v in hoisted_prop_vars]
                        inner_lookups.append(f.logical(prop_lookups, hoisted_props))

            # Create the arguments for constructing the compound key. Only include the
            # values for non-null keys. Null key names are still included.
            values: list[ir.Value] = [compound_key.type]
            for key in info.original_keys:
                assert isinstance(key.type, ir.ScalarType)
                values.append(ir.Literal(types.String, key.type.name))
                if key in curr_non_null_keys:
                    values.append(key)
            inner_lookups.append(ir.Construct(
                None,
                tuple(values),
                compound_key,
                OrderedSet().frozen()
            ))

            # Add the output for this combination of null/non-null keys.
            # First, get the annos
            annos = list(info.original_output.annotations)

            if info.original_output.keys:
                annos.append(f.annotation(builtins.output_keys, tuple(info.original_output.keys)))

            # Then, construct the output aliases. Drop the arguments that are either
            # properties of currently null keys or the null keys themselves
            curr_null_keys_or_properties = OrderedSet[ir.Var]()
            output_vars = helpers.output_vars(info.original_output.aliases)
            for key in curr_null_keys:
                for tasks in info.inner_property_lookups.get(key, []):
                    for task in tasks:
                        assert isinstance(task, ir.Lookup)
                        vars = helpers.vars(task.args)
                        curr_null_keys_or_properties.update([v for v in vars if v in output_vars])
                if key in output_vars:
                    curr_null_keys_or_properties.add(key)

            new_output_aliases = []
            for (alias, var) in info.original_output.aliases:
                # Need a placeholder (None) for this output var so that columns still
                # line up. This will be excluded later during Flatten so we don't
                # actually output None values.
                out_var = None if var in curr_null_keys_or_properties else var
                new_output_aliases.append((alias, out_var))
            inner_lookups.append(f.output(new_output_aliases, [compound_key], annos=annos))

            return [f.logical(inner_lookups, [])]

    # Null keys should not be looked up. Non-nullable keys should also not be looked up,
    # since they will be looked up separately in their own logical.
    #
    # Returns the removed lookups as a dict mapping from key to lookups
    def _remove_from_key_lookups(self, lookups: list[ir.Task], vars_to_purge: OrderedSet[ir.Var], non_nullable_keys: OrderedSet[ir.Var]):
        removed = defaultdict(list)
        # keep track of transitive keys too
        keys = [[v] for v in vars_to_purge]

        there_is_progress = True
        while there_is_progress:
            there_is_progress = False
            for task in lookups:
                assert isinstance(task, ir.Lookup)
                vars = helpers.vars(task.args)
                for ks in keys:
                    if vars[-1] in ks:
                        lookups.remove(task)
                        # ks[0] is the original key that led to this chain of lookups
                        removed[ks[0]].append(task)
                        new_vars = [v for v in vars if v not in vars_to_purge and v not in non_nullable_keys]
                        ks.extend(new_vars)
                        there_is_progress = True
                        break

        return removed



import warnings

from relationalai.semantics.reasoners.graph.paths.path_algorithms.single import single_shortest_path, single_walk

__all__ = ["single_shortest_path", "single_walk"]

warnings.warn(
    "relationalai.early_access.paths.path_algorithms.single is deprecated. "
    "Please migrate to relationalai.semantics.reasoners.graph.paths.path_algorithms.single",
    DeprecationWarning,
    stacklevel=2,
)


import warnings

from relationalai.semantics.lqp.executor import LQPExecutor

__all__ = ["LQPExecutor"]

warnings.warn(
    "relationalai.early_access.lqp.executor is deprecated, "
    "Please migrate to relationalai.semantics.lqp.executor",
    DeprecationWarning,
    stacklevel=2,
)


from relationalai.semantics import Model, Integer, define, select
from relationalai.semantics.reasoners.graph.paths.graph import Graph
from relationalai.semantics.reasoners.graph.paths.path_algorithms.single import single_shortest_path


model = Model("test_single_paths", dry_run=False)


# First test with grid graph
grid = Graph.construct_grid(model, 3)

n = Integer.ref()
u = grid.Node.ref()

n_1_1 = grid.Node.new(row = 1, col = 1)
n_1_2 = grid.Node.new(row = 1, col = 2)
n_1_3 = grid.Node.new(row = 1, col = 3)
n_2_1 = grid.Node.new(row = 2, col = 1)
n_2_2 = grid.Node.new(row = 2, col = 2)
n_2_3 = grid.Node.new(row = 2, col = 3)
n_3_1 = grid.Node.new(row = 3, col = 1)
n_3_2 = grid.Node.new(row = 3, col = 2)
n_3_3 = grid.Node.new(row = 3, col = 3)

Source_1 = model.Concept("Source_1", extends=[grid.Node])
Target_1 = model.Concept("Target_1", extends=[grid.Node])

define(Source_1(n_1_1))
define(Target_1(n_2_2))

path_1 = single_shortest_path(grid, Source_1, Target_1)
iter_path_1 = select(n, u.row, u.col).where(path_1(n, u)).to_df()
set_path_1 = set(row for row in iter_path_1.itertuples(index = False, name = None))

expected_path_1_1 = {
    (0, 1, 1), (1, 1, 2), (2, 2, 2)
}

expected_path_1_2 = {
    (0, 1, 1), (1, 2, 1), (2, 2, 2)
}

assert(
    set_path_1 == expected_path_1_1 or
    set_path_1 == expected_path_1_2
)


# Second test with grid graph
Target_2 = model.Concept("Target_2", extends=[grid.Node])

define(Target_2(n_3_3))

path_2 = single_shortest_path(grid, Source_1, Target_2)
iter_path_2 = select(n, u.row, u.col).where(path_2(n, u)).to_df()
set_path_2 = set(row for row in iter_path_2.itertuples(index = False, name = None))

expected_path_2_1 = {
    (0, 1, 1), (1, 2, 1), (2, 3, 1), (3, 3, 2), (4, 3, 3)
}

expected_path_2_2 = {
    (0, 1, 1), (1, 2, 1), (2, 2, 2), (3, 3, 2), (4, 3, 3)
}

expected_path_2_3 = {
    (0, 1, 1), (1, 2, 1), (2, 2, 2), (3, 2, 3), (4, 3, 3)
}

expected_path_2_4 = {
    (0, 1, 1), (1, 1, 2), (2, 2, 2), (3, 3, 2), (4, 3, 3)
}

expected_path_2_5 = {
    (0, 1, 1), (1, 1, 2), (2, 2, 2), (3, 2, 3), (4, 3, 3)
}

expected_path_2_6 = {
    (0, 1, 1), (1, 1, 2), (2, 1, 3), (3, 2, 3), (4, 3, 3)
}


assert(
    set_path_2 == expected_path_2_1 or
    set_path_2 == expected_path_2_2 or
    set_path_2 == expected_path_2_3 or
    set_path_2 == expected_path_2_4 or
    set_path_2 == expected_path_2_5 or
    set_path_2 == expected_path_2_6
)


# Third test with grid graph
Source_3 = model.Concept("Source_3", extends=[grid.Node])
Target_3 = model.Concept("Target_3", extends=[grid.Node])

define(Source_3(n_1_1))
define(Source_3(n_1_2))
define(Target_3(n_3_2))
define(Target_3(n_3_3))

path_3 = single_shortest_path(grid, Source_3, Target_3)
iter_path_3 = select(n, u.row, u.col).where(path_3(n, u)).to_df()
set_path_3 = set(row for row in iter_path_3.itertuples(index = False, name = None))

expected_path_3 = {
    (0, 1, 2), (1, 2, 2), (2, 3, 2)
}

assert(
    set_path_3 == expected_path_3
)


# Fourth test with grid graph
Source_4 = model.Concept("Source_4", extends=[grid.Node])
Target_4 = model.Concept("Target_4", extends=[grid.Node])

define(Source_4(n_1_1))
define(Source_4(n_2_2))
define(Target_4(n_2_2))
define(Target_4(n_3_3))

path_4 = single_shortest_path(grid, Source_4, Target_4)
iter_path_4 = select(n, u.row, u.col).where(path_4(n, u)).to_df()
set_path_4 = set(row for row in iter_path_4.itertuples(index = False, name = None))

expected_path_4 = {
    (0, 2, 2)
}

assert(
    set_path_4 == expected_path_4
)


# Fifth test with grid graph
Target_5 = model.Concept("Target_5", extends=[grid.Node])

n_4_4 = grid.Node.new(row = 4, col = 4)

define(Target_5(n_4_4))

path_5 = single_shortest_path(grid, Source_1, Target_5)
iter_path_5 = select(n, u.row, u.col).where(path_5(n, u)).to_df()
set_path_5 = set(row for row in iter_path_5.itertuples(index = False, name = None))

expected_path_5 = set()

assert(
    set_path_5 == expected_path_5
)



import warnings

from relationalai.semantics.reasoners.graph.paths.path_algorithms.find_paths import find_shortest_paths, find_walks

__all__ = ["find_shortest_paths", "find_walks"]

warnings.warn(
    "relationalai.early_access.paths.path_algorithms.find_paths is deprecated. "
    "Please migrate to relationalai.semantics.reasoners.graph.paths.path_algorithms.find_paths",
    DeprecationWarning,
    stacklevel=2,
)


from .splinter import Splinter
from .list_types import RewriteListTypes
from .gc_nodes import GarbageCollectNodes
from .flatten import Flatten
from .dnf_union_splitter import DNFUnionSplitter
from .extract_keys import ExtractKeys
from .extract_nested_logicals import ExtractNestedLogicals
from .fd_constraints import FDConstraints

__all__ = ["Splinter", "RewriteListTypes", "GarbageCollectNodes", "Flatten", "DNFUnionSplitter", "ExtractKeys", "ExtractNestedLogicals", "FDConstraints"]



import warnings

from relationalai.semantics.reasoners.graph.paths.path_algorithms.one_sided_ball_upto import ball_upto

__all__ = ["ball_upto"]

warnings.warn(
    "relationalai.early_access.paths.path_algorithms.one_sided_ball_upto is deprecated. "
    "Please migrate to relationalai.semantics.reasoners.graph.paths.path_algorithms.one_sided_ball_upto",
    DeprecationWarning,
    stacklevel=2,
)